"""Train random forests."""
import matplotlib.pyplot as plt
import numpy as np
from sklearn import ensemble, metrics, tree
from sklearn.datasets import fetch_california_housing
from sklearn.model_selection import train_test_split


def train_dt_and_rf(
    x_train: np.ndarray, x_test: np.ndarray, y_train: np.ndarray, y_test: np.ndarray
) -> dict[str, list]:
    """Train decision trees and random forests for regression.

    Define a decision tree regressor with max_depth parameter value of 1.
    Train it and compute the MSE on the test set. Repeat these steps for all max_depth
    values between 2 and 30 (including 30). Afterwards, perform all these operations
    with a random forest regressor.

    Args:
        x_train (np.ndarray): The training data.
        x_test (np.ndarray): The test data.
        y_train (np.ndarray): The training labels.
        y_test (np.ndarray): The test labels.

    Returns:
        dict[str, list]: Two lists containing the MSEs for the decision tree and the random forest regressors.
    """
    # 3. create and fit a DT regressor
    decision_tree = tree.DecisionTreeRegressor(max_depth=1)
    decision_tree.fit(x_train, y_train)

    # 4. make predictions on test data and compute and print MSE
    predictions = decision_tree.predict(x_test)
    mse = metrics.mean_squared_error(y_test, predictions)
    print("MSE =", mse)

    # 5. repeat steps 3 and 4 for higher max_depth values
    depths = np.arange(1, 31)
    decision_trees_mse = [mse]

    for max_depth in depths[1:]:
        decision_tree = tree.DecisionTreeRegressor(max_depth=max_depth)
        decision_tree.fit(x_train, y_train)
        predictions = decision_tree.predict(x_test)
        decision_trees_mse.append(metrics.mean_squared_error(y_test, predictions))

    # 6. repeat steps 3,4 and 5 with random forests
    random_forests_mse = []

    for max_depth in depths:
        random_forest = ensemble.RandomForestRegressor(
            max_depth=max_depth, n_estimators=10
        )
        random_forest.fit(x_train, y_train)
        predictions = random_forest.predict(x_test)
        random_forests_mse.append(metrics.mean_squared_error(y_test, predictions))

    # 7. return both MSE lists
    return {
        "decision_trees_mse": decision_trees_mse,  # list with MSEs of decision trees
        "random_forests_mse": random_forests_mse,  # list with MSEs of random forests
    }


if __name__ == "__main__":
    np.random.seed(0)
    # 1. import dataset
    california_housing = fetch_california_housing()

    # 2. split dataset into train and test
    train_data, test_data, train_targets, test_targets = train_test_split(
        california_housing.data,
        california_housing.target,
        test_size=0.1,
        random_state=21,
    )

    # 8. call `train_dt_and_rf` function and get MSE curves
    result_dict = train_dt_and_rf(train_data, test_data, train_targets, test_targets)
    dt_mse = result_dict["decision_trees_mse"]
    rf_mse = result_dict["random_forests_mse"]

    # 9. plot both MSE curves
    depths = np.arange(1, 31)
    plt.plot(depths, dt_mse, label="decision tree")
    plt.plot(depths, rf_mse, label="random forest")

    plt.title("MSE comparison: decision trees vs random forests")
    plt.xlabel("max tree depth")
    plt.ylabel("mean squared error")
    plt.legend()
    plt.savefig("./figures/mse_random_forest.png")
    plt.show()

    # 10. Observations:
    # MSE initially drops when increasing the max_depth, but starts
    # to increase again after max_depth=10. This is due to overfitting on the
    # training data.

    # MSE of the random forests steadily decreases and is smaller
    # overall compared to the single decision trees. This shows that the ensemble
    # approach provides an advantage in scenarios where overfitting can occur.
